@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2016 Codership Oy <info@codership.com>
+ * Copyright (C) 2008-2019 Codership Oy <info@codership.com>
  *
  * $Id$
  */
@@ -51,6 +51,9 @@ gcs_group_init (gcs_group_t* group, gcache_t* const cache,
     group->prim_seqno = GCS_SEQNO_ILL;
     group->prim_num   = 0;
     group->prim_state = GCS_NODE_STATE_NON_PRIM;
+    group->prim_gcs_ver  = 0;
+    group->prim_repl_ver = 0;
+    group->prim_appl_ver = 0;
 
     *(gcs_proto_t*)&group->gcs_proto_ver = gcs_proto_ver;
     *(int*)&group->repl_proto_ver = repl_proto_ver;
@@ -231,6 +234,32 @@ group_go_non_primary (gcs_group_t* group)
     // what else? Do we want to change anything about the node here?
 }
 
+static void
+group_check_proto_ver(gcs_group_t* group)
+{
+    assert(group->quorum.primary); // must be called only on primary CC
+
+    gcs_node_t& node(group->nodes[group->my_idx]);
+    bool fail(false);
+
+#define GROUP_CHECK_NODE_PROTO_VER(LEVEL)                               \
+    if (node.LEVEL < group->quorum.LEVEL) {                             \
+        gu_fatal("Group requested %s: %d, max supported by this node: %d." \
+                 "Upgrade the node before joining this group."          \
+                 "Need to abort.",                                      \
+                 #LEVEL, group->quorum.LEVEL, node.LEVEL);              \
+        fail = true;                                                    \
+    }
+
+    GROUP_CHECK_NODE_PROTO_VER(gcs_proto_ver);
+    GROUP_CHECK_NODE_PROTO_VER(repl_proto_ver);
+    GROUP_CHECK_NODE_PROTO_VER(appl_proto_ver);
+
+#undef GROUP_CHECK_NODE_PROTO_VER
+
+    if (fail) gu_abort();
+}
+
 static const char group_empty_id[GCS_COMP_MEMB_ID_MAX_LEN + 1] = { 0, };
 
 static void
@@ -342,6 +371,14 @@ group_post_state_exchange (gcs_group_t* group)
         }
 
         assert (group->prim_num > 0);
+
+#define GROUP_UPDATE_PROTO_VER(LEVEL) \
+        if (group->prim_##LEVEL##_ver < quorum->LEVEL##_proto_ver) \
+            group->prim_##LEVEL##_ver = quorum->LEVEL##_proto_ver;
+        GROUP_UPDATE_PROTO_VER(gcs);
+        GROUP_UPDATE_PROTO_VER(repl);
+        GROUP_UPDATE_PROTO_VER(appl);
+#undef GROUP_UPDATE_PROTO_VER
     }
     else {
         // non-primary configuration
@@ -367,6 +404,7 @@ group_post_state_exchange (gcs_group_t* group)
              quorum->appl_proto_ver,
              GU_UUID_ARGS(&quorum->group_uuid));
 
+    if (quorum->primary) group_check_proto_ver(group);
     group_check_donor(group);
 }
 
@@ -474,8 +512,8 @@ gcs_group_handle_comp_msg (gcs_group_t* group, const gcs_comp_msg_t* comp)
                 /* initialize node ID to the one given by the backend - this way
                  * we'll be recognized as coming from prev. conf. in node array
                  * remap below */
-                strncpy ((char*)group->nodes[0].id, new_nodes[0].id,
-                         sizeof (new_nodes[0].id) - 1);
+                strncpy (group->nodes[0].id, new_nodes[0].id,
+                         sizeof (group->nodes[0].id));
                 group->nodes[0].segment = new_nodes[0].segment;
             }
         }
@@ -747,20 +785,20 @@ gcs_group_handle_join_msg  (gcs_group_t* group, const gcs_recv_msg_t* msg)
             }
         }
         else {
-            if (sender_idx == peer_idx) {
-                if (GCS_NODE_STATE_JOINED == sender->status) {
-                    gu_info ("Member %d.%d (%s) resyncs itself to group",
-                             sender_idx, sender->segment, sender->name);
+            if (GCS_NODE_STATE_JOINED == sender->status) {
+                if (sender_idx == peer_idx) {
+                    gu_info("Member %d.%d (%s) resyncs itself to group.",
+                            sender_idx, sender->segment, sender->name);
                 }
                 else {
-                    assert(sender->desync_count > 0);
-                    return 0; // don't deliver up
+                    gu_info("%d.%d (%s): State transfer %s %d.%d (%s) complete.",
+                            sender_idx, sender->segment, sender->name, st_dir,
+                            peer_idx, peer ? peer->segment : -1, peer_name);
                 }
             }
             else {
-                gu_info ("%d.%d (%s): State transfer %s %d.%d (%s) complete.",
-                         sender_idx, sender->segment, sender->name, st_dir,
-                         peer_idx, peer ? peer->segment : -1, peer_name);
+                assert(sender->desync_count > 0);
+                return 0; // don't deliver up
             }
         }
     }
@@ -1528,6 +1566,9 @@ group_get_node_state (const gcs_group_t* const group, long const node_idx)
         node->gcs_proto_ver,
         node->repl_proto_ver,
         node->appl_proto_ver,
+        group->prim_gcs_ver,
+        group->prim_repl_ver,
+        group->prim_appl_ver,
         node->desync_count,
         flags
         );
